/*
+-+-+-+-+-+-+-+-+-
*/
/*
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):
** Author, Date:	Aniruddha Gokhale, 31-July-1995
** Intent:		Header file for the IODV facility
** Last Update:		$Author: rwalkup $, $Date: 1997/02/18 17:04:35 $
** Source File:		$RCSfile: dicom_iodv.h,v $
** Revision:		$Revision: 1.16 $
** Status:		$State: Exp $
*/

#ifndef _DICOM_IODV_H_
#define _DICOM_IODV_H_

#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#ifdef SOLARIS
#include <sys/fcntl.h>
#endif
#include <sys/stat.h>
#include <ctype.h>
#include <setjmp.h>
#include "dicom.h"
#include "condition.h"
#include "lst.h"
#include "dicom_objects.h"
#include "dicom_uids.h"

typedef enum {
    IODV_K_MANDATORY,
    IODV_K_CONDITIONAL,
    IODV_K_OPTIONAL
}   IODV_MODULEREQUIREMENT;

typedef enum {
    IODV_K_TYPE1,
    IODV_K_TYPE1C,
    IODV_K_TYPE2,
    IODV_K_TYPE2C,
    IODV_K_TYPE3
}   IODV_ATTRIBUTEREQUIREMENT;

typedef enum {
    IODV_K_UNKNOWNIOD,
    IODV_K_CT,
    IODV_K_CR,
    IODV_K_MR,
    IODV_K_US,
    IODV_K_USMF,
    IODV_K_SC,
    IODV_K_NM,
}   IODV_IODTYPE;

typedef enum {
    IODV_K_UNKNOWNIE,
    IODV_K_PATIENTIE,
    IODV_K_STUDYIE,
    IODV_K_SERIESIE,
    IODV_K_FRAMEOFREFERENCEIE,
    IODV_K_EQUIPMENTIE,
    IODV_K_IMAGEIE,
    IODV_K_OVERLAYIE,
    IODV_K_CURVEIE
}   IODV_IETYPE;

typedef enum {
    IODV_K_UNKNOWNMODULE,
    IODV_K_PATIENTMODULE,
    IODV_K_GENERALSTUDYMODULE,
    IODV_K_PATIENTSTUDYMODULE,
    IODV_K_GENERALSERIESMODULE,
    IODV_K_CRSERIESMODULE,
    IODV_K_USSERIESMODULE,
    IODV_K_NMSERIESMODULE,
    IODV_K_FRAMEOFREFERENCEMODULE,
    IODV_K_USFRAMEOFREFERENCEMODULE,
    IODV_K_GENERALEQUIPMENTMODULE,
    IODV_K_SCEQUIPMENTMODULE,
    IODV_K_SOPCOMMONMODULE,
    IODV_K_GENERALIMAGEMODULE,
    IODV_K_IMAGEPLANEMODULE,
    IODV_K_IMAGEPIXELMODULE,
    IODV_K_CINEMODULE,
    IODV_K_MULTIFRAMEMODULE,
    IODV_K_CONTRASTMODULE,
    IODV_K_CRIMAGEMODULE,
    IODV_K_CTIMAGEMODULE,
    IODV_K_MRIMAGEMODULE,
    IODV_K_NMIMAGEMODULE,
    IODV_K_USIMAGEMODULE,
    IODV_K_USMULTIFRAMEIMAGEMODULE,
    IODV_K_USREGIONCALIBRATIONIMAGEMODULE,
    IODV_K_SCIMAGEMODULE,
    IODV_K_OVERLAYPLANEMODULE,
    IODV_K_LOOKUPTABLEMODULE,
    IODV_K_CURVEMODULE,
    IODV_K_OVERLAYIDENTIFICATIONMODULE,
    IODV_K_CURVEIDENTIFICATIONMODULE,
    IODV_K_CURVEPLANEMODULE,
    IODV_K_NMIMAGEPIXELMODULE,
    IODV_K_NMMULTIFRAMEIMAGEMODULE,
    IODV_K_NMISOTOPEIMAGEMODULE,
    IODV_K_NMDETECTORIMAGEMODULE,
    IODV_K_NMTOMOACQIMAGEMODULE,
    IODV_K_NMMULTIGATEDACQIMAGEMODULE,
    IODV_K_NMPHASEIMAGEMODULE,
    IODV_K_NMRECONSTRUCTIONIMAGEMODULE,
    IODV_K_MULTIFRAMEOVERLAYPLANEMODULE,
    IODV_K_AUDIOMODULE,
    IODV_K_VOILUTIMAGEMODULE,
    IODV_K_MODALITYLUTMODULE,
    IODV_K_LUTIDENTIFICATION
}   IODV_MODULETYPE;

/* define the type for IODV_STATUS */
#if LONGSIZE == 64
typedef unsigned int IODV_STATUS;
#define IODV_MAKESTATUS(p, s) (((((unsigned int)(p)) << 16) & 0xffff0000) \
| ((unsigned int)(s) & 0xffff))
#else
typedef unsigned long IODV_STATUS;
#define IODV_MAKESTATUS(p, s) (((((unsigned long)(p)) << 16) & 0xffff0000) \
| ((unsigned long)(s) & 0xffff))
#endif
#define IODV_STATUS_PRIMARY(s) (unsigned short) ((s) >> 16)
#define IODV_STATUS_SECONDARY(s) (unsigned short) ((s) & 0xffff)

/* define all the PRIMARY status conditions that may result after
   verification */
 /* conditions applicable to modules and attributes */
#define IODV_K_DONTCARE		 0x0	/* ignore any status */
#define IODV_K_GOOD		 (1 << 0)	/* satisfies all constraints */
#define IODV_K_ABSENT		 (1 << 1)	/* absent when it was
						 * REQUIRED */
#define IODV_K_EXTRA		 (1 << 2)	/* Present even when not reqd */

 /* conditions applicable only to attributes */
#define IODV_K_COND1FAILED	 (1 << 3)	/* condition1 failed */
#define IODV_K_COND2FAILED	 (1 << 4)	/* condition2 failed */
#define IODV_K_RESTRICT1FAILED   (1 << 5)	/* restriction1 failed */
#define IODV_K_RESTRICT2FAILED   (1 << 6)	/* restriction2 failed */
#define IODV_K_UNKNOWNCONDFAILED (1 << 7)	/* unknown condition failed */
#define IODV_K_VMFAILED		 (1 << 8)	/* value multiplicity error */
#define IODV_K_VRFAILED		 (1 << 9)	/* value representation error */
#define IODV_K_ABSENTINOPTMODULE (1 << 10)	/* absent, but module is
						 * optional */
#define IODV_K_ABSENTBUTUNKNOWN  (1 << 11)	/* absent, but attr.
						 * condition is unknown */

 /* conditions applicable to modules only */
#define IODV_K_MODCONDFAILED	 (1 << 3)	/* module condition failed */
#define IODV_K_MODATTRERROR 	 (1 << 4)	/* some or all REQD attr
						 * absent or some error */
/* define various SECONDARY status codes */
/* the value 0x0 is once again don't care */
#define IODV_K_NOTVERIFIED	(1 << 0)	/* was not verified */
#define IODV_K_NOTPRESENT	(1 << 1)	/* attr not present */
#define IODV_K_EXTRA		(1 << 2)	/* attr is extra */
 /* these are for Bad Value Multiplicity */
#define IODV_K_ZEROLENGTH	(1 << 3)	/* 0 length type 1 attr */
#define IODV_K_BADLENGTH	(1 << 4)	/* bad length */
 /* these are for bad VR */
#define IODV_K_UNEVENLENGTH	(1 << 5)	/* str uneven in length */
#define IODV_K_INVALIDCHAR	(1 << 6)	/* invalid characters in data */
#define IODV_K_BADTRAILINGCHARS	(1 << 7)	/* trailing chars bad */
 /* these are for conditional type 2 */
#define IODV_K_ELEMNOTFOUND	(1 << 8)	/* TAG not present */
#define IODV_K_VALUNMATCHED	(1 << 9)	/* val not present in list */

typedef struct {
    LST_HEAD *iodList;
}   IODV_IODLIST;

typedef struct {
    void *reserved[2];
    IODV_IODTYPE iodType;
    char classuid[DICOM_UI_LENGTH + 1];	/* IOD SOP class */
    char *name;			/* IOD name */
    LST_HEAD *ieList;		/* list of IEs */
}   IODV_IODDESCRIPTION;

typedef struct {
    void *reserved[2];
    char *name;
    IODV_IETYPE ieType;		/* IE type */
    LST_HEAD *modList;		/* list of modules that make up ths IE */
}   IODV_IEDESCRIPTION;

typedef struct {
    void *reserved[2];
    char *name;
    IODV_MODULETYPE modType;	/* Module type */
    IODV_MODULEREQUIREMENT modReq;	/* Module usage */
    CTNBOOLEAN modCheck;
    LST_HEAD *attrList;		/* list of attributes */
    IODV_STATUS status;		/* status of verification */
}   IODV_MODDESCRIPTION;

typedef struct {
    CTNBOOLEAN range;		/* whether its a range or not */
    char lower;			/* lower, default 1 when no range */
    char upper;			/* a number or alphabet 'n' */
}   IODV_ATTR_VM;

typedef DCM_TAG IODV_TYPE_COND1;/* condition1 is just a tag */

typedef struct {
    /*
     * cond2 is made up of a single tag followed by list of string values
     */
    DCM_TAG tag;
    int index;
    LST_HEAD *strList;
}   IODV_TYPE_COND2;

typedef struct {
    CTNBOOLEAN mult;
    LST_HEAD *strList;
}   IODV_TYPE_RESTRICT1;

typedef struct {
    CTNBOOLEAN mult;
    LST_HEAD *numList;
}   IODV_TYPE_RESTRICT2;

typedef CTNBOOLEAN IODV_TYPE_UNKNOWN;

typedef struct {
    IODV_TYPE_COND1 cond1;	/* condition 1 */
    IODV_TYPE_COND2 *cond2;	/* condition2 */
    IODV_TYPE_RESTRICT1 *restrict1;	/* restriction type 1 */
    IODV_TYPE_RESTRICT2 *restrict2;	/* restriction type 2 */
    IODV_TYPE_UNKNOWN unknown;	/* is it an unknown cond */
}   IODV_ATTRVERIFY;

typedef struct {
    IODV_ATTR_VM *vm;		/* value multiplicity, default 1 */
    IODV_ATTRVERIFY *verify;	/* various verification elements */
    LST_HEAD *seq;		/* sequence */
}   IODV_OPTATTRFIELDS;

typedef struct {
    void *reserved[2];
    DCM_TAG tag;		/* DICOM tag */
    IODV_ATTRIBUTEREQUIREMENT attrReq;	/* Attr Usage */
    IODV_OPTATTRFIELDS *opt;	/* all the rest of the optional fields */
    IODV_STATUS status;		/* status of verification */
}   IODV_ATTRDESCRIPTION;

/* used when the restriction1 is a multilist of string values ,
   Also used to maintain list of elements of a DCM_SEQUENCE_ITEM
*/
typedef struct {
    void *reserved[2];
    LST_HEAD *l;
}   IODV_MULT_LIST;

/* used to represent list of strings in restriction1 */
typedef struct {
    void *reserved[2];
    char *str;
}   IODV_ENUMERATED_STRING_VALUES;

/* used to represent a list of integers in restriction2 */
typedef struct {
    void *reserved[2];
    int number;
}   IODV_ENUMERATED_INT_VALUES;

/* define a struture that holds information related to a given status code */
typedef struct {
    int count;			/* number of strings being held */
    char *info[20];		/* pointers to the various strings */
}   IODV_STATUSMEANING;

/* define a structure that holds all attributes that are EXTRA in the object */
typedef struct {
    int count;
    DCM_TAG *tag;
}   IODV_EXTRA_ATTR;

#define IODV_NORMAL	/* NORMAL return */ \
	FORM_COND(FAC_IODV, SEV_SUCC, 1)
#define IODV_LISTCREATEFAILED	/* LST_Create failed */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 2)
#define IODV_MALLOCFAILED	/* out of memory */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 3)
#define IODV_LISTENQUEUEFAILED	/* enqueue into list failed */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 4)
#define IODV_PARSEFAILED	/* failed to parse the file */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 5)
#define IODV_FILEOPENFAILED	/* failed to open parse file */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 6)
#define IODV_SYNTAXERROR	/* syntax error in parse file */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 7)
#define IODV_LISTDESTROYFAILED	/* failed to destroy a list */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 8)
#define IODV_FREEFAILED	/* failed to free up memory */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 9)
#define IODV_DUMPFAILED	/* failed to dump info */\
	FORM_COND(FAC_IODV, SEV_ERROR, 10)
#define IODV_NULLOBJECT	/* Null object for verification */\
	FORM_COND(FAC_IODV, SEV_ERROR, 11)
#define IODV_NULLIODLIST /* Null iod list for verification */\
	FORM_COND(FAC_IODV, SEV_ERROR, 12)
#define IODV_DCMIMGTYPENOTFOUND	/* failure to find object type */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 13)
#define IODV_IODINFONOTPRESENT	/* info for this IOD not present */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 14)
#define IODV_ELEMLISTCREATEFAILED  /* failed to create element list */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 15)
#define IODV_ELEMLISTDESTROYFAILED  /* failed to destroy element list */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 16)
#define IODV_MODULEVERIFYFAILED  /* failed to verify module */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 17)
#define IODV_ATTRIBUTEVERIFYFAILED  /* failed to verify attribute */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 18)
#define IODV_CONDUNSATISFIED  /* condition unsatisfied */ \
	FORM_COND(FAC_IODV, SEV_WARN, 19)
#define IODV_DCMFACFAILED  /* DCM facility failed */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 20)
#define IODV_COND1VRFYFAILED  /* failed to verify cond 1 */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 21)
#define IODV_COND2VRFYFAILED  /* failed to verify cond 2 */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 22)
#define IODV_RESTRICT1VRFYFAILED  /*  failed to vrfy restriction 1 */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 23)
#define IODV_RESTRICT2VRFYFAILED  /* failed to verify restriction 2*/ \
	FORM_COND(FAC_IODV, SEV_ERROR, 24)
#define IODV_VMVRFYFAILED  /* failed to verify value multiplicity */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 25)
#define IODV_VRVRFYFAILED  /* failed to verify value representtaion */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 26)
#define IODV_UNKNOWNVRFYFAILED  /* failed to verify unknown condition */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 27)
#define IODV_SEQINFOABSENT  /* sequence information is absent in the list */\
	FORM_COND(FAC_IODV, SEV_ERROR, 28)
#define IODV_SEQVRFYFAILED  /* sequence verification failed */\
	FORM_COND(FAC_IODV, SEV_ERROR, 29)
#define IODV_IEVRFYFAILED  /* IE verification failed */\
	FORM_COND(FAC_IODV, SEV_ERROR, 30)
#define IODV_NULLIOD /* Null iod passed for verification */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 31)
#define IODV_VERIFYFAILED /* Overall verification failed */ \
	FORM_COND(FAC_IODV, SEV_ERROR, 32)

char *IODV_Message(CONDITION);
CONDITION IODV_ParseFile(char *fname, IODV_IODLIST ** iodl);
CONDITION IODV_Free(IODV_IODLIST **);
CONDITION IODV_DumpIODList(IODV_IODLIST **, FILE *);
CONDITION IODV_DumpIOD(IODV_IODDESCRIPTION *, FILE *, CTNBOOLEAN);
CONDITION IODV_Verify(DCM_OBJECT **, IODV_IODLIST **, IODV_IODDESCRIPTION **);
IODV_STATUSMEANING *IODV_GetModuleStatusMeaning(IODV_STATUS status);
IODV_STATUSMEANING *IODV_GetAttributeStatusMeaning(IODV_STATUS status, 
						   CTNBOOLEAN *statusFlag);
#endif
